from ajson.json_type_reports import JsonTypeReports
from ajson.type_inspector import TypeInspector


def AJson():
    def wrapper(cls: type):
        report = TypeInspector().inspect_type(cls)
        JsonTypeReports().add(cls, report)
        return cls

    return wrapper


import json
from inspect import isfunction
from typing import Dict, List, Optional, Set, Type

from typeguard import check_type

from ajson.singleton import Singleton

ISO_FORMAT = '%Y-%m-%dT%H:%M:%S.%f'


class AJsonAnnotationParseError(Exception):
    pass


class AJsonValidationError(TypeError):
    pass


class AJsonEmptyRequiredAttributeError(AJsonValidationError):
    pass


class _AttrReport(object):
    def __init__(self, attribute_name: str, hint: Optional[Type], **kwargs):
        self.groups: Optional[Set[str]] = kwargs.get('groups', None)
        self.name: str = kwargs.get('name', attribute_name)
        self.datetime_format: str = kwargs.get('d_format', ISO_FORMAT)  # iso format
        self.attribute_name: str = attribute_name
        self.required: str = kwargs.get('required', 'false')

        self.hint: Type = hint

        if self.groups is not None:
            self.groups = set(json.loads(self.groups))
            for group in self.groups:
                if not isinstance(group, (str, int)):
                    raise AJsonAnnotationParseError('unable to parse groups for attribute "{}"'.format(attribute_name))
        self.required = self.required.lower() != 'false'


class _TypeReport(object):
    def __init__(self, attr_reports: Dict[str, _AttrReport], hint: Type):
        self.report_map: Dict[str, _AttrReport] = attr_reports
        self.hint = hint

    def get(self, attr_name: str) -> _AttrReport:
        return self.report_map.get(attr_name, _AttrReport(attr_name, None))

    def get_attribute_names(self, groups: Optional[List[str]] = None) -> Optional[List[str]]:
        if groups is None:
            return None
        return [
            key for key, report in self.report_map.items()
            if report.groups is not None and
               len(report.groups.intersection(groups)) > 0
        ]

    def get_by_serialize_name_or_default(self, name: str) -> Optional[_AttrReport]:
        try:
            return next(x for x in self.report_map.values() if x.name == name)
        except StopIteration as e:
            if name in self.report_map:
                # if there is a report for that variable but not with that name, ignore it
                raise e
            # if there is no conflict, create a new basic __AttrReport in case the entity got that attr dynamically
            return _AttrReport(name, hint=None)

    # todo move this function to the serializer, it doesn't make sense to have this in a report
    def validate_instance(self, instance):
        for key, report in self.report_map.items():
            if hasattr(instance, report.attribute_name) and getattr(instance, report.attribute_name) is not None:
                value = getattr(instance, report.attribute_name)
                if report.hint is not None:
                    try:
                        check_type(report.attribute_name, value, report.hint)
                    except TypeError:
                        error_text = 'attribute: "{0}.{1}" do not have type: {2}'
                        error_text_params = self.hint.__name__, report.attribute_name, report.hint
                        raise AJsonValidationError(error_text.format(*error_text_params))

            elif report.required:
                error_text = 'required attribute: "{0}.{1}" is empty'
                raise AJsonEmptyRequiredAttributeError(error_text.format(self.hint.__name__, report.attribute_name))


class JsonTypeReports(object, metaclass=Singleton):
    def __init__(self):
        self.reports: Dict[type, _TypeReport] = {}

    def add(self, _type: Type, type_report_dict: Dict):
        type_report = {}
        for key, attribute_report in type_report_dict.items():
            type_report[key] = _AttrReport(key, hint=None, **attribute_report)

        # merge parent report with the new one
        for parent_class in _type.__bases__:
            if parent_class in self.reports:
                type_report = {**self.reports[parent_class].report_map, **type_report}

        # adding extra reports for the attributes that have a type but not a @aj annotation
        if hasattr(_type, '__annotations__'):
            for key, attr_hint in _type.__annotations__.items():
                type_report[key] = type_report.get(key, _AttrReport(key, hint=None))
                type_report[key].hint = attr_hint

        # adding extra reports for the for the properties that don't have @aj annotation
        for key, value in vars(_type).items():
            if key.startswith('_'):
                continue
            if isinstance(value, property):
                type_report[key] = type_report.get(key, _AttrReport(key, hint=None))
                if hasattr(value.fget, '__annotations__'):
                    type_report[key].hint = value.fget.__annotations__
            elif not isfunction(value):
                type_report[key] = type_report.get(key, _AttrReport(key, hint=None))
                if hasattr(_type, '__annotations__') and _type.__annotations__.get(key, False):
                    type_report[key].hint = _type.__annotations__[key]
        self.reports[_type] = _TypeReport(type_report, _type)

    def clear(self):
        self.reports = {}



from ajson.regex import as_comment_regex
from ajson.singleton import Singleton


class CommentHandler(object, metaclass=Singleton):
    def format_class_source(self, source: str) -> str:
        single = source.find("'''")
        double = source.find('"""')

        single = single if single != -1 else len(source) + 1
        double = double if double != -1 else len(source) + 1

        first_pos = min(single, double)
        if first_pos == len(source) + 1:
            return source

        if single == first_pos:
            end_of_comment = source.find("'''", first_pos + 3) + 3
            source = self._add_as_comment_as_inline(source, first_pos, end_of_comment)

        if double == first_pos:
            end_of_comment = source.find('"""', first_pos + 3) + 3
            source = self._add_as_comment_as_inline(source, first_pos, end_of_comment)

        return self.format_class_source(source)

    def _add_as_comment_as_inline(self, source: str, first_pos: int, end_of_comment: int):
        comment = source[first_pos: end_of_comment]
        as_comments = as_comment_regex.findall(comment)
        if len(as_comments) >= 1:
            source = source[:first_pos] + source[end_of_comment:]
            previous_line_pos = source[:first_pos].rfind("\n")
            source = source[:previous_line_pos] + "  # " + as_comments[0] + source[previous_line_pos:]
        else:
            source = source[:first_pos] + source[end_of_comment:]
        return source





import unittest

from ajson.class_decorator import AJson
from ajson.json_type_reports import JsonTypeReports, ISO_FORMAT


class TestAJsonDecorator(unittest.TestCase):
    def setUp(self):
        self.old_reports = JsonTypeReports().reports
        JsonTypeReports().clear()

    def tearDown(self):
        JsonTypeReports().reports = self.old_reports

    def test_annotation_class_creates_report_with_group(self):
        @AJson()
        class AJDA:
            def __init__(self):
                self.a = 10  # testing @aj(groups=["admin"])

        reports = JsonTypeReports().reports
        self.assertIn(AJDA, reports)
        self.assertEqual(len(reports.keys()), 1)
        self.assertEqual(reports[AJDA].get("a").groups, {"admin"})
        self.assertEqual(reports[AJDA].get("a").datetime_format, ISO_FORMAT)
        self.assertEqual(reports[AJDA].get("a").name, "a")

    def test_annotation_class_creates_report_with_group_and_name(self):
        @AJson()
        class AJDB:
            def __init__(self):
                self.a = 10  # testing @aj(groups=["admin"] name=annotation)

        reports = JsonTypeReports().reports
        self.assertIn(AJDB, reports)
        self.assertEqual(reports[AJDB].get("a").datetime_format, ISO_FORMAT)
        self.assertEqual(reports[AJDB].get("a").name, "annotation")

    def test_annotation_class_creates_report_all_parameters(self):
        @AJson()
        class AJDC:
            def __init__(self):
                self.a = 10
                '''
                    @aj( 
                    groups="[
                        'admin',
                        'public'
                        ]"
                    name=annotation
                    d_format=Y-M-D
                    )
                '''

        reports = JsonTypeReports().reports
        self.assertIn(AJDC, reports)
        self.assertEqual(reports[AJDC].get("a").groups, {"admin", "public"})
        self.assertEqual(reports[AJDC].get("a").datetime_format, "Y-M-D")
        self.assertEqual(reports[AJDC].get("a").name, "annotation")

    def test_class_with_parent_get_parent_properties(self):
        @AJson()
        class AIPP1:
            a: int = 10  # @aj(name=aa)
            b: int = 20  # @aj(name=bb)

        @AJson()
        class AIPP2(AIPP1):
            c: int = 30  # @aj(name=cc)
            d: int = 40  # @aj(name=dd)

        reports = JsonTypeReports().reports
        self.assertIn(AIPP1, reports)
        self.assertIn(AIPP2, reports)
        self.assertEqual(reports[AIPP1].get("a").name, 'aa')
        self.assertEqual(reports[AIPP1].get("b").name, 'bb')
        self.assertEqual(reports[AIPP2].get("a").name, 'aa')
        self.assertEqual(reports[AIPP2].get("b").name, 'bb')
        self.assertEqual(reports[AIPP2].get("c").name, 'cc')
        self.assertEqual(reports[AIPP2].get("d").name, 'dd')

    def test_class_with_parent_overwrites_parent_properties(self):
        @AJson()
        class AIWP1:
            a: int = 10  # @aj(name=aa)

        @AJson()
        class AIWP2(AIWP1):
            a: int = 30  # @aj(name=aaa)
            d: int = 40  # @aj(name=dd)

        reports = JsonTypeReports().reports
        self.assertIn(AIWP1, reports)
        self.assertIn(AIWP2, reports)
        self.assertEqual(reports[AIWP1].get("a").name, 'aa')
        self.assertEqual(reports[AIWP2].get("a").name, 'aaa')
        self.assertEqual(reports[AIWP2].get("d").name, 'dd')



import json
import unittest

from ajson.type_inspector import TypeInspector


class TestClassInspector(unittest.TestCase):
    def test_simple_class_with_inline_annotation_gets_the_as_params_and_identifies_the_attributes(self):
        class CIA:
            def __init__(self):
                self.a = 10  # testing @aj(groups=["test"] required)
                self.b = 10  # @ac( "groups": ["test"])

        report = TypeInspector().inspect_type(CIA)
        self.assertEqual(len(report.keys()), 1)
        self.assertDictEqual(report, {
            'a': {'groups': '["test"]', 'required': ''}
        })

    def test_parse_report_with_as_in_multiline(self):
        class CIB:
            def __init__(self):
                self.a = 10
                '''
                    @aj(
                        groups="[
                            1,
                            2,
                            3
                        ]"
                    )
                '''

        report = TypeInspector().inspect_type(CIB)
        self.assertEqual(len(report.keys()), 1)
        self.assertEqual(json.loads(report['a']['groups']), [1, 2, 3])

    def test_annotation_without_attribute_is_ignore(self):
        class CID:
            def __init__(self):
                '''
                    @aj(
                        "groups":  1
                    )
                '''
                self.b = 6  # @aj(name=test)

        report = TypeInspector().inspect_type(CID)
        self.assertEqual(len(report.keys()), 1)
        self.assertDictEqual(report, {
            'b': {'name': 'test'}
        })

    def test_annotation_gets_properties_too(self):
        class CIEP:
            def __init__(self):
                self._a = 10

            @property
            def a(self):
                """ @aj(name=new_a) """
                return self._a + 10

        report = TypeInspector().inspect_type(CIEP)
        self.assertEqual(len(report.keys()), 1)
        self.assertDictEqual(report, {
            'a': {'name': 'new_a'}
        })



import inspect
import unittest

from ajson.comment_handler import CommentHandler


class TestCommentHandler(unittest.TestCase):
    def test_clears_single_comments(self):
        class CHB:
            def __init__(self):
                self.a = 2
                '''
                    this is a test
                '''
                self.b = 2
                '''
                    another test
                '''

        cleared_class = CommentHandler().format_class_source(inspect.getsource(CHB))

        self.assertEqual(cleared_class, """        class CHB:
            def __init__(self):
                self.a = 2
                
                self.b = 2
                \n""")

    def test_clears_double_comments(self):
        class CHC:
            def __init__(self):
                self.a = 3
                """
                    this is a test
                """
                self.b = 3
                """
                    another test
                """

        cleared_class = CommentHandler().format_class_source(inspect.getsource(CHC))

        self.assertEqual(cleared_class, """        class CHC:
            def __init__(self):
                self.a = 3
                
                self.b = 3
                \n""")

    def test_move_as_comment_to_inline_in_single(self):
        class CHD:
            def __init__(self):
                self.a = 4
                '''
                    @aj(group=2)
                '''

        cleared_class = CommentHandler().format_class_source(inspect.getsource(CHD))

        self.assertEqual(cleared_class, """        class CHD:
            def __init__(self):
                self.a = 4  # @aj(group=2)
                
""")

    def test_complex_case(self):
        class CHE:
            def __init__(self):
                self.a = 4
                '''
                    @aj(group=0)
                '''
                self.b = 8  # ANOTHER TEST @aj(group=1)
                self.c = self.b
                """
                    this is a comment and then we have the as_comment 
                    __example__ @aj(group=2)
                """

        cleared_class = CommentHandler().format_class_source(inspect.getsource(CHE))

        self.assertEqual(cleared_class, """        class CHE:
            def __init__(self):
                self.a = 4  # @aj(group=0)
                
                self.b = 8  # ANOTHER TEST @aj(group=1)
                self.c = self.b  # @aj(group=2)
                
""")



import inspect
import logging
import os
from typing import AnyStr, Dict, Match, Optional, Type

from bs4 import BeautifulSoup

from ajson.comment_handler import CommentHandler
from ajson.regex import as_comment_regex, find_attribute_regex
from ajson.singleton import Singleton


class TypeInspector(object, metaclass=Singleton):

    def inspect_type(self, _type: Type) -> Dict[str, Dict]:
        source = inspect.getsource(_type)
        clean_source = CommentHandler().format_class_source(source)

        return {
            **self._get_aj_annotation_reports(clean_source),
            **self._get_properties_report(_type)
        }

    def _get_aj_annotation_reports(self, source) -> Dict[str, Dict]:
        report = {}

        while True:
            matches = as_comment_regex.search(source)
            if matches is None:
                break
            try:
                aj_dict = self.aj_str_to_aj_dict(matches)
                attribute = self._find_attribute(source, matches.start())
                if attribute is not None:
                    report[attribute] = aj_dict
            except Exception as e:
                logging.warning("Unable to parse @aj {}".format(matches[0]))
            source = source[matches.end():]

        return report

    def _get_properties_report(self, _type: Type) -> Dict[str, Dict]:
        properties = [(key, value) for key, value in vars(_type).items() if
                      isinstance(value, property) and value.__doc__]

        report = {}
        for name, prop in properties:
            matches = as_comment_regex.search(prop.__doc__)
            if matches is None:
                break
            aj_dict = self.aj_str_to_aj_dict(matches)
            report[name] = aj_dict
        return report

    def _find_attribute(self, source, as_position) -> Optional[str]:
        source_to_as = source[:as_position]
        line: str = source_to_as[source_to_as.rfind("\n"):].strip()
        attributes_found = find_attribute_regex.findall(line)
        if len(attributes_found) >= 1:
            if attributes_found[0][1] == '':
                return attributes_found[0][0][5:]
            else:
                return attributes_found[0][1]

        return None

    def aj_str_to_aj_dict(self, matches: Match[AnyStr]) -> Optional[Dict]:
        # getting the aj string and removing the # in case the json is multi line
        aj_str = matches.string[matches.start(): matches.end()].replace("#", "")
        xml_aj_str = "<tag {} ></tag>".format(aj_str[4:-1].replace(os.linesep, ''))
        aj_params_dict = BeautifulSoup(xml_aj_str, features="html.parser").find('tag').attrs
        return {k: v.replace("'", '"') for k, v in aj_params_dict.items()}



import collections
import json
from datetime import datetime
from typing import Any, Callable, Dict, List, NewType, Optional, Set, Tuple, Type, Union

from ajson.json_type_reports import ISO_FORMAT, JsonTypeReports, _AttrReport, _TypeReport

Groups = NewType('Groups', Optional[List[str]])
Handler = NewType('Handler', Callable[[Any, Groups, _AttrReport], Any])


class ASerializer:
    """
    Serialize and unserialize objects
    """
    max_depth: int
    """
    Defines how many nested objects should be serialized.
    If the object reaches to this point, the result will be replaced by "..."

    >>> serializer = ASerializer(max_depth=2)
    >>> nested_dict = {"d1": {"d2": {"d3": "value deep inside"}}}
    >>> serializer.serialize(nested_dict)
    "{"d1": {"d2": "..." }}"
    """
    def __init__(self, max_depth=15):
        self.max_depth: int = max_depth
        self._handlers: Dict[Type, Handler] = {}

    def add_serialize_handler(self, _type: Type, handler: Handler):
        """
        Adds a handler for a specific type to modify the way it should be serialize

        >>> serializer = ASerializer()
        >>> serializer.add_serialize_handler(int, lambda obj, *args: obj if obj > 0 else 0) # negative ints return 0
        >>> serializer.serialize(5)
        '5'
        >>> serializer.serialize(-6)
        '0'
        """
        self._handlers[_type] = handler

    def serialize(self, obj, groups: Optional[List[str]] = None) -> str:
        """
        Creates a json string from the obj

        :param obj: Object to be serialize
        :param groups: list of groups that determines what attributes should be serialize

        >>> from ajson import AJson
        >>> serializer = ASerializer()
        >>> @AJson()
        ... class House:
        ...    rooms_num: int  # @aj(groups='["public", "owner"]')
        ...    square_meters: int  # @aj(groups='["owner"]')
        ...    def __init__(self, rooms_num, square_meters):
        ...        self.rooms_num = rooms_num
        ...        self.square_meters = square_meters

        >>> serializer.serialize(House(3, 100), groups=['public'])
        '{"room_num": 3}'
        >>> serializer.serialize(House(3, 100), groups=['owner'])
        '{"room_num": 3, "square_meters": 100}'

        """
        return json.dumps(self._to_dict_recursive(obj, groups, 0))

    def to_dict(self, obj, groups: Optional[List[str]] = None) -> Union[Dict[str, Any], List]:
        """
        Same as serialize, but it creates a serializable dict instead of a str

        :param obj: Object to be serialize
        :param groups: list of groups that determines what attributes should be serialize

        >>> from ajson import AJson
        >>> serializer = ASerializer()
        >>> @AJson()
        ... class Car:
        ...    max_speed: float  # @aj(groups='["basic", "detailed"]')
        ...    brand: str  # @aj(groups='["detailed"]')
        ...    def __init__(self, max_speed, brand):
        ...        self.max_speed = max_speed
        ...        self.brand = brand

        >>> serializer.to_dict(Car(140, 'ford'), groups=['basic'])
        {"max_speed": 140}
        >>> serializer.to_dict(Car(140, 'ford'), groups=['detailed'])
        {"max_speed": 140, "brand": 7}

        """
        return self._to_dict_recursive(obj, groups, 0)

    def _to_dict_recursive(self, obj, groups: Optional[List[str]] = None, depth=0, attr_report: _AttrReport = None):
        depth += 1
        if depth > self.max_depth:
            return '...'
        for class_ in self._handlers:
            if isinstance(obj, class_):
                return self._handlers[class_](obj, groups, attr_report)
        if obj is None:
            return None
        elif isinstance(obj, (int, str, float)):
            return obj
        elif isinstance(obj, datetime):
            return self.__datetime_handler(obj, attr_report)
        elif isinstance(obj, (list, tuple, set)):
            return self.__list_handler(obj, groups, depth)
        elif isinstance(obj, dict):
            return self.__dict_handler(obj, groups, depth)
        else:
            return self.__object_handler(obj, groups, depth)

    def __list_handler(self, obj: list, groups: Optional[List[str]], depth: int):
        serialized_list = []
        obj = list(obj)
        for item in obj:
            serialized_list.append(self._to_dict_recursive(item, groups=groups, depth=depth))
        return serialized_list

    def __dict_handler(self,
                       obj: dict,
                       groups: Optional[List[str]],
                       depth,
                       class_report: Optional[_TypeReport] = None
                       ):
        serialized_dict = {}
        for key, value in obj.items():
            # we don't want to serialize private attributes if we don't have a class report
            if class_report is not None or not str(key).startswith('_'):
                if class_report is None:
                    attr_report = None
                else:
                    attr_report = class_report.get(key)
                    key = attr_report.name
                serialized_dict[key] = self._to_dict_recursive(value, groups, depth, attr_report)
        return serialized_dict

    def __datetime_handler(self, obj: datetime, attr_report: Optional[_AttrReport] = None) -> str:
        if attr_report is None:
            return obj.isoformat()

        return obj.strftime(attr_report.datetime_format)

    def __object_handler(self, obj: object, groups: Optional[List[str]], depth):
        class_report = JsonTypeReports().reports.get(obj.__class__, None)
        if class_report is None:
            attributes = {key: value for key, value in obj.__dict__.items()
                          if not isinstance(value, collections.Callable) and
                          not 'key'.startswith('_')}
            return self.__dict_handler(attributes, groups, depth)
        attributes_to_serialize = class_report.get_attribute_names(groups)
        if attributes_to_serialize is None:
            attributes_to_serialize = (k for k in (*obj.__dict__.keys(), *obj.__class__.__dict__.keys()) if
                                       not k.startswith('_'))

        attributes = {key: getattr(obj, key) for key in attributes_to_serialize}
        return self.__dict_handler(attributes, groups, depth, class_report)

    def unserialize(self, json_str: str, _type: Optional[Type] = None, *init_args_array, **init_kargs) -> Any:
        """
        Creates an object with the type `_type` from a string
        groups will be ignored for unserialization

        :param json_str: string to be transformed into an object
        :param _type: Resulting type of the object to construct
        :param init_args_array: construct args list to initialize the object with type `_type`
        :param init_kargs: construct args to initialize the object with type `_type`

        >>> from ajson import AJson
        >>> serializer = ASerializer()
        >>> @AJson()
        ... class House:
        ...    rooms_num: int  # @aj()
        ...    square_meters: int  # @aj()

        >>> house: House = serializer.unserialize('{"rooms_num": 1, "square_meters":50}', House)
        >>> house.rooms_num
        1
        >>> house.square_meters
        50
        """
        return self.from_dict(json.loads(json_str), _type, *init_args_array, **init_kargs)

    def from_dict(self, dict_obj: Any, _type: Optional[Type] = None, *init_args_array, **init_kargs) -> Any:
        """
        Creates an object with the type `_type` from a dictionary
        groups will be ignored for unserialization

        :param dict_obj: dict to be transformed into an object
        :param _type: Resulting type of the object to construct
        :param init_args_array: construct args list to initialize the object with type `_type`
        :param init_kargs: construct args to initialize the object with type `_type`

        >>> from ajson import AJson
        >>> serializer = ASerializer()
        >>> @AJson()
        ... class Car:
        ...    max_speed: float  # @aj(')
        ...    brand: str  # @aj()

        >>> car: Car = serializer.from_dict({'max_speed': 100, 'brand': 'Jeep'}, Car)
        >>> car.max_speed
        100
        >>> car.brand
        'Jeep'
        """

        return self._from_dict_recursive(dict_obj, _type, *init_args_array, **init_kargs)

    def _from_dict_recursive(self, dict_obj: Any, _type: Optional[Type] = None, attr_report: _AttrReport = None,
                             *init_args_array, **init_kargs) -> Any:
        if isinstance(dict_obj, (list, tuple, set)):
            return self._unserialize_list(_type, dict_obj, init_args_array, init_kargs)
        elif isinstance(dict_obj, dict):
            return self._unserialize_obj(_type, dict_obj, init_args_array, init_kargs)
        elif isinstance(dict_obj, str):
            return self._unserialize_str_or_date(attr_report, dict_obj)

        return dict_obj

    def _unserialize_obj(self, _type: Optional[Type], dict_obj: Any, init_args_array, init_kargs):
        type_report = JsonTypeReports().reports.get(_type, None)
        if type_report is None or _type is None:
            return {k: self._from_dict_recursive(v) for k, v in dict_obj.items()}
        result_obj = _type(*init_args_array, **init_kargs)
        for key, value in dict_obj.items():
            try:
                attr_report = type_report.get_by_serialize_name_or_default(key)
            except StopIteration:
                # do nothing if report is not found
                continue
            result_dict = self._from_dict_recursive(value, _type=attr_report.hint, attr_report=attr_report)
            if hasattr(result_obj, attr_report.attribute_name) or attr_report.hint is not None:
                setattr(result_obj, attr_report.attribute_name, result_dict)

        type_report.validate_instance(result_obj)
        return result_obj

    def _unserialize_str_or_date(self, attr_report: _AttrReport, dict_obj: Any) -> Any:
        # check if it is a date time
        if attr_report is not None:
            datetime_format = attr_report.datetime_format
        else:
            datetime_format = ISO_FORMAT
        try:
            return datetime.strptime(dict_obj, datetime_format)
        except ValueError:
            return dict_obj

    def _unserialize_list(self, _type: Optional[Type], dict_obj: Any, init_args_array, init_kargs) -> Any:
        if _type is None:
            return [self._from_dict_recursive(item, _type, *init_args_array, **init_kargs) for item in dict_obj]
        list_type = None
        if _type is not None and getattr(_type, '__args__', None) is not None and len(_type.__args__) > 0:
            list_type = _type.__args__[0]

        if issubclass(_type, List):
            return [self._from_dict_recursive(item, list_type, *init_args_array, **init_kargs) for item in dict_obj]
        elif issubclass(_type, Set):
            return {self._from_dict_recursive(item, list_type, *init_args_array, **init_kargs) for item in dict_obj}
        elif issubclass(_type, Tuple):
            return (self._from_dict_recursive(item, list_type, *init_args_array, **init_kargs) for item in dict_obj)
        else:
            return [self._from_dict_recursive(item, _type, *init_args_array, **init_kargs) for item in dict_obj]



import json
import unittest
from datetime import timedelta, datetime

from ajson.aserializer import ASerializer
from ajson.json_type_reports import JsonTypeReports, ISO_FORMAT


class TestSerialization(unittest.TestCase):
    def setUp(self):
        self.serializer = ASerializer()
        self.old_reports = JsonTypeReports().reports
        JsonTypeReports().clear()

    def tearDown(self):
        JsonTypeReports().reports = self.old_reports

    def test_basic_object_serialization(self):
        serialization = self.serializer.serialize(5)
        self.assertTrue(serialization == "5", "number serialization")
        serialization = self.serializer.serialize("hi")
        self.assertTrue(serialization == '"hi"', "str serialization")

    def test_simple_objects(self):
        class SSimpleObject(object):
            def __init__(self):
                self.a = 1
                self.b = "hi"
                self.c = None

        obj_dict = self.serializer.to_dict(SSimpleObject())
        self.assertTrue(obj_dict["a"] == 1)
        self.assertTrue(obj_dict["b"] == "hi")
        self.assertIsNone(obj_dict["c"])

    def test_complex_objects(self):
        class SComplexObject(object):
            def __init__(self):
                self.a = {"a": 10, 1: 15}
                self.b = [1, 2, "hello"]

        serialization = self.serializer.serialize(SComplexObject())
        ser_obj = json.loads(serialization)
        self.assertTrue(isinstance(ser_obj["a"], dict))
        self.assertTrue(ser_obj["a"]["a"] == 10)
        self.assertTrue(ser_obj["a"]["1"] == 15)
        self.assertTrue(len(ser_obj["b"]) == 3)

    def test_cycle_ref(self):
        class SCycleRefObject(object):
            def __init__(self):
                self.a = self

        serialization = self.serializer.serialize(SCycleRefObject())
        ser_obj = json.loads(serialization)
        checking_obj = ser_obj
        for i in range(self.serializer.max_depth):
            self.assertIn("a", checking_obj)
            checking_obj = checking_obj["a"]
        self.assertNotIn("a", checking_obj)

    def test_date_time_objects(self):
        date = datetime.now()
        date_dict = self.serializer.to_dict({"datetime": date})
        parsed_datetime = datetime.strptime(date_dict["datetime"], ISO_FORMAT)
        self.assertTrue(
            datetime.now() - parsed_datetime < timedelta(minutes=5),
            datetime.now() - parsed_datetime
        )

    def test_handler_modify_specific_types(self):
        self.serializer.add_serialize_handler(datetime, lambda d, g, a: "test")
        date = datetime.now()
        date_dict = self.serializer.to_dict({"datetime": date})
        self.assertEqual(date_dict["datetime"], "test")

    # Unserialize

    def test_simple_unserialize_returns_dict(self):
        class SSimpleObject(object):
            def __init__(self):
                self.a = 5
                self.b = "hi"
                self.c = None

        obj_src = self.serializer.serialize(SSimpleObject())
        obj_dict = self.serializer.unserialize(obj_src)
        self.assertEqual(obj_dict["a"], 5)
        self.assertEqual(obj_dict["b"], "hi")
        self.assertIsNone(obj_dict["c"])

    def test_simple_unserialize_date_time_with_iso_format(self):
        date = datetime.now()
        date_src = self.serializer.serialize({"datetime": date})
        date_dict = self.serializer.unserialize(date_src)
        self.assertIsInstance(date_dict["datetime"], datetime)



from datetime import datetime
from typing import List, Optional, Union

from ajson import AJson


@AJson()
class SEmptyObjectAJson(object):
    def __init__(self):
        self.a = 1


@AJson()
class SSimpleObjectAJson(object):
    def __init__(self):
        self.a = 1  # @aj()
        self.b = 1


@AJson()
class SSimpleObjectAJsonWithGroups(object):
    def __init__(self):
        self.a = 1  # @aj(groups=["admin"])
        self.b = 2  # @aj(groups=["public"])


@AJson()
class SObjectWithGroupsAndNoGroups(object):
    def __init__(self):
        self.a = 1  # @aj(groups=["admin"])
        self.b = 2  # @aj(groups=["public"])
        self.c = 8  # @aj()
        self.d = 9


@AJson()
class SObjectWithProperties(object):
    def __init__(self):
        self._a = 1
        self._b = 2

    @property
    def a(self):
        """ @aj() """
        return self._a

    @property
    def b(self):
        """ @aj(groups='["g1"]') """
        return self._b


@AJson()
class SSimpleObjectAJsonNested1(object):
    def __init__(self):
        self.a = 1  # @aj(groups=["admin"])
        self.b = 2  # @aj(groups=["public"])


@AJson()
class SSimpleObjectAJsonNested2(object):
    def __init__(self):
        self.nested1 = SSimpleObjectAJsonNested1()  # @aj(groups=["admin"])
        self.nested2 = SSimpleObjectAJsonNested1()  # @aj(groups=["public"])


@AJson()
class SSimpleObjectWithDate(object):
    def __init__(self):
        self.time1 = datetime(2000, 2, 1, 5, 30)  # @aj(d_format="%Y/%m/%d")
        self.time2 = datetime(2010, 5, 10, 2, 40)  # @aj(d_format="%Y--%H%M")


@AJson()
class USSimpleObjectAJson(object):
    def __init__(self):
        self.a = 1


@AJson()
class USNameAndDateObjectAJson(object):
    a = 1  # @aj(name="my_mane")
    date = None  # @aj(d_format="%Y--%H%M")


@AJson()
class USNameAndDateObjectAJson2(object):
    a = 1  # @aj(name="my_mane")
    date: datetime = None  # @aj(d_format="%Y/%m/%d")


@AJson()
class USNestedObject0(object):
    a: int  # @aj(name="my_mane")
    date: datetime  # @aj(d_format="%Y/%m/%d")

    def __init__(self):
        self.a = 1
        self.date: datetime = None


@AJson()
class USNestedObject1(object):
    nested: USNestedObject0 = USNestedObject0()  # @aj()


@AJson()
class USNestedListObject(object):
    nested_list: List[USNestedObject0] = [USNestedObject0()]  # @aj()


@AJson()
class USRequiredObject(object):
    a: int  # @aj(required)
    b: int  # @aj()


@AJson()
class USWithoutAnnotationsObject(object):
    a: int
    b: int


@AJson()
class USWithHintsObject(object):
    a: int
    b: str


@AJson()
class USWithOptionalHintsObject(object):
    a: Optional[int]
    b: str


@AJson()
class USWithMultiTypeHintsObject(object):
    a: Union[int, str]


@AJson()
class USWithProperties(object):
    def __init__(self):
        self._a = 1
        self._b = 2

    @property
    def a(self) -> int:
        """ @aj() """
        return self._a

    @property
    def b(self):
        """ @aj(name=new_b) """
        return self._b

    @a.setter
    def a(self, a):
        self._a = a

    @b.setter
    def b(self, b):
        self._b = b

    def this_is_a_func(self) -> str:
        return "testing"


def div(a: int, b: int) -> float:
    """Divide a by b"""
    return a / b



import json
import unittest

from ajson.aserializer import ASerializer
from ajson.json_type_reports import AJsonEmptyRequiredAttributeError, AJsonValidationError
from tests.types_for_tests.test_serializaer_with_annotations_types import *


class TestSerializationWithAnnotations(unittest.TestCase):
    # Test with AJson decoration
    def setUp(self):
        self.serializer = ASerializer()

    def test_empty_object_with_ajson_serialize_obj_if_no_groups_passed(self):
        serialization = self.serializer.serialize(SEmptyObjectAJson())
        self.assertEqual(len(json.loads(serialization).keys()), 1)
        self.assertEqual(json.loads(serialization)['a'], 1)

        serialization = self.serializer.serialize(SEmptyObjectAJson(), groups=[])
        self.assertEqual(len(json.loads(serialization).keys()), 0)

    def test_object_with_empty_aj_annotation_is_serialized_too_if_no_group_is_passed(self):
        serialization = self.serializer.serialize(SSimpleObjectAJson())
        self.assertEqual(len(json.loads(serialization).keys()), 2)
        self.assertEqual(json.loads(serialization)['a'], 1)
        self.assertEqual(json.loads(serialization)['b'], 1)

        serialization = self.serializer.serialize(SSimpleObjectAJson(), groups=[])
        self.assertEqual(len(json.loads(serialization).keys()), 0)

    def test_object_with_ajson_returns_only_attributes_in_group(self):
        serialization = self.serializer.serialize(SSimpleObjectAJsonWithGroups(), groups=['admin'])
        self.assertEqual(len(json.loads(serialization).keys()), 1)
        self.assertEqual(json.loads(serialization)['a'], 1)

        serialization = self.serializer.serialize(SSimpleObjectAJsonWithGroups(), groups=['public'])
        self.assertEqual(len(json.loads(serialization).keys()), 1)
        self.assertEqual(json.loads(serialization)['b'], 2)

    def test_object_with_ajson_returns_only_attributes_in_group_with_nested_references(self):
        dict_obj = self.serializer.to_dict(SSimpleObjectAJsonNested2(), groups=['admin'])
        self.assertEqual(len(dict_obj.keys()), 1)
        self.assertEqual(dict_obj['nested1'], {'a': 1})

        dict_obj = self.serializer.to_dict(SSimpleObjectAJsonNested2(), groups=['admin', 'public'])
        self.assertEqual(len(dict_obj.keys()), 2)
        self.assertEqual(dict_obj['nested1'], {'a': 1, 'b': 2})
        self.assertEqual(dict_obj['nested2'], {'a': 1, 'b': 2})

    def test_object_with_ajson_and_date_format_returns_the_right_date_format(self):
        dict_obj = self.serializer.to_dict(SSimpleObjectWithDate())
        self.assertEqual(len(dict_obj.keys()), 2)
        self.assertEqual(dict_obj['time1'], '2000/02/01')
        self.assertEqual(dict_obj['time2'], '2010--0240')

    def test_list_ob_objects_is_serialized_with_the_right_groups(self):
        list_obj = self.serializer.to_dict([SSimpleObjectAJsonWithGroups()], groups=['admin'])
        self.assertEqual(len(list_obj), 1)
        self.assertEqual(list_obj[0]['a'], 1)

    def test_serialize_only_attrs_with_the_right_groups(self):
        obj_dict = self.serializer.to_dict(SObjectWithGroupsAndNoGroups(), groups=['admin'])
        self.assertEqual(len(obj_dict.keys()), 1)
        self.assertEqual(obj_dict['a'], 1)

        obj_dict = self.serializer.to_dict(SObjectWithGroupsAndNoGroups(), groups=['public'])
        self.assertEqual(len(obj_dict.keys()), 1)
        self.assertEqual(obj_dict['b'], 2)

        obj_dict = self.serializer.to_dict(SObjectWithGroupsAndNoGroups())
        self.assertEqual(len(obj_dict.keys()), 4)

    def test_serialize_properties_too(self):
        obj_dict = self.serializer.to_dict(SObjectWithProperties())
        self.assertEqual(len(obj_dict.keys()), 2)
        self.assertEqual(obj_dict['a'], 1)
        self.assertEqual(obj_dict['b'], 2)

        obj_dict = self.serializer.to_dict(SObjectWithProperties(), groups=['g1'])
        self.assertEqual(len(obj_dict.keys()), 1)
        self.assertEqual(obj_dict['b'], 2)

    # Unserialize
    def test_simple_entity_is_unserialize_from_dict(self):
        dict_obj = {
            'a': 10
        }

        obj: USSimpleObjectAJson = self.serializer.from_dict(dict_obj, USSimpleObjectAJson)
        self.assertEqual(obj.a, 10)

    def test_unserialize_with_different_name_nade_date_time_format(self):
        dict_obj = {
            'a': 10,  # this should be ignored
            'my_mane': 20,
            'date': '2010--0240',
        }

        obj: USNameAndDateObjectAJson = self.serializer.from_dict(dict_obj, USNameAndDateObjectAJson)
        self.assertEqual(obj.a, 20)
        self.assertIsInstance(obj.date, datetime)

    def test_unserialize_list_returns_list_of_objects(self):
        list_dict = [
            {
                'my_mane': 1,
                'date': '2000/01/01',
            },
            {
                'my_mane': 2,
                'date': '2000/02/01',
            },
            {
                'my_mane': 3,
                'date': '2000/03/01',
            },
        ]

        list_obj: List[USNameAndDateObjectAJson2] = self.serializer.from_dict(list_dict, USNameAndDateObjectAJson2)
        self.assertEqual(len(list_obj), 3)
        self.assertEqual(list_obj[0].a, 1)
        self.assertEqual(list_obj[0].date.month, 1)
        self.assertEqual(list_obj[1].date.month, 2)
        self.assertEqual(list_obj[2].date.month, 3)

    def test_unserialize_nested_objects(self):
        obj_dict = {
            'nested': {
                'my_mane': 3,
                'date': '2003/01/01',
            },
        }

        obj: USNestedObject1 = self.serializer.from_dict(obj_dict, USNestedObject1)
        self.assertEqual(obj.nested.a, 3)
        self.assertIsInstance(obj.nested.date, datetime)
        self.assertEqual(obj.nested.date.year, 2003)

    def test_unserialize_nested_with_object_lists(self):
        obj_dict = {
            'nested_list': [
                {
                    'my_mane': 3,
                    'date': '2003/01/01',
                }
            ],
        }

        obj: USNestedListObject = self.serializer.from_dict(obj_dict, USNestedListObject)
        self.assertEqual(obj.nested_list[0].a, 3)
        self.assertIsInstance(obj.nested_list[0].date, datetime)
        self.assertEqual(obj.nested_list[0].date.year, 2003)

    def test_unserialize_with_required_throws_if_required_is_not_provided(self):
        dict_obj = {
            'b': 1,
        }
        with self.assertRaises(AJsonEmptyRequiredAttributeError):
            self.serializer.from_dict(dict_obj, USRequiredObject)

    def test_unserialize_with_required_throws_if_required_is_null(self):
        str_obj = '{"b": 1, "a": null }'
        with self.assertRaises(AJsonEmptyRequiredAttributeError):
            self.serializer.unserialize(str_obj, USRequiredObject)

    def test_unserialize_object_without_annotations_should_be_generated_with_default_behaviour(self):
        dict_obj = {
            'b': 1,
            'a': 10,
        }
        obj: USWithoutAnnotationsObject = self.serializer.from_dict(dict_obj, USWithoutAnnotationsObject)
        self.assertEqual(obj.a, 10)
        self.assertEqual(obj.b, 1)

    def test_unserialize_raises_error_if_type_do_not_match(self):
        dict_obj = {
            'b': None,
            'a': 10,
        }
        obj: USWithHintsObject = self.serializer.from_dict(dict_obj, USWithHintsObject)

        self.assertIsNone(obj.b)
        self.assertEqual(obj.a, 10)

    def test_unserialize_allows_validation_with_multi_type(self):
        obj: USWithMultiTypeHintsObject = self.serializer.from_dict({'a': 10}, USWithMultiTypeHintsObject)
        self.assertEqual(obj.a, 10)
        obj: USWithMultiTypeHintsObject = self.serializer.from_dict({'a': 'str'}, USWithMultiTypeHintsObject)
        self.assertEqual(obj.a, 'str')
        with self.assertRaises(AJsonValidationError):
            self.serializer.from_dict({'a': 10.13}, USWithMultiTypeHintsObject)

    def test_unserialize_properties(self):
        obj: USWithProperties = self.serializer.from_dict({'a': 10, 'b': 5}, USWithProperties)
        self.assertEqual(obj.a, 10)
        self.assertEqual(obj.b, 2)  # not updated as it is not using the name
        obj: USWithProperties = self.serializer.from_dict({'a': 4, 'new_b': 20}, USWithProperties)
        self.assertEqual(obj.a, 4)
        self.assertEqual(obj.b, 20)



# -*- coding: utf-8 -*-

"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# To use a consistent encoding
from codecs import open
from os import path

# Always prefer setuptools over distutils
from setuptools import setup, find_packages

import ajson

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
try:
    with open(path.join(here, 'README.md')) as f:
        long_description = f.read()
except:
    long_description = 'ajson'

setup(
    name='ajson',
    version=ajson.__version__,

    description='simple serializer based on annotations',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/JorgeGarciaIrazabal/ajson',

    author='Jorge Garcia Irazabal',
    author_email='jorge.girazabal@gmail.com',

    # Choose your license
    license='GNU General Public License v3 (GPLv3)',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='json serializer annotation validation',
    packages=find_packages(exclude="_static"),
    install_requires=[
        'beautifulsoup4',
        'typeguard >= 2.2.0, <2.3.0'
    ],
    test_suite="tests",
)